@@ -19,16 +19,25 @@ def ocr_finalize(*pargs, **kwargs):
         dafilelist = kwargs['dafilelist']
         filename = kwargs['filename']
         file_list = []
+        input_number = target.number
         for parg in pargs:
             if type(parg) is list:
                 for item in parg:
                     if type(item) is ReturnValue:
-                        file_list.append(item.value['doc']._pdf_page_path(int(item.value['page'])))
+                        if isinstance(item.value, dict):
+                            if 'page' in item.value:
+                                file_list.append([item.value['indexno'], int(item.value['page']), item.value['doc']._pdf_page_path(int(item.value['page']))])
+                            else:
+                                file_list.append([item.value['indexno'], 0, item.value['doc'].path()])
             else:
                 if type(parg) is ReturnValue:
-                    file_list.append(parg.value['doc']._pdf_page_path(int(parg.value['page'])))
+                    if isinstance(item.value, dict):
+                        if 'page' in item.value:
+                            file_list.append([parg.value['indexno'], int(parg.value['page']), parg.value['doc']._pdf_page_path(int(parg.value['page']))])
+                        else:
+                            file_list.append([parg.value['indexno'], 0, parg.value['doc'].path()])
         from docassemble.base.pandoc import concatenate_files
-        pdf_path = concatenate_files(file_list)
+        pdf_path = concatenate_files([y[2] for y in sorted(file_list, key=lambda x: x[0]*10000 + x[1])])
         target.initialize(filename=filename, extension='pdf', mimetype='application/pdf', reinitialize=True)
         shutil.copyfile(pdf_path, target.file_info['path'])
         del target.file_info
@@ -43,10 +52,10 @@ def ocr_finalize(*pargs, **kwargs):
         if type(parg) is list:
             for item in parg:
                 #sys.stderr.write("ocr_finalize: sub item is a " + str(type(item)) + "\n")
-                if type(item) is ReturnValue:
+                if type(item) is ReturnValue and isinstance(item.value, dict):
                     output[int(item.value['page'])] = item.value['text']
         else:
-            if type(parg) is ReturnValue:
+            if type(parg) is ReturnValue and isinstance(item.value, dict):
                 output[int(parg.value['page'])] = parg.value['text']
         #index += 1
     #sys.stderr.write("ocr_finalize: assembling output\n")
@@ -142,12 +151,17 @@ def ocr_page_tasks(image_file, language=None, psm=6, x=None, y=None, W=None, H=N
     todo = list()
     for doc in image_file:
         if hasattr(doc, 'extension'):
-            if doc.extension not in ['pdf', 'png', 'jpg', 'gif']:
+            if doc.extension not in ['pdf', 'png', 'jpg', 'gif', 'docx', 'doc', 'odt', 'rtf']:
                 raise Exception("document with extension " + doc.extension + " is not a readable image file")
             if doc.extension == 'pdf':
                 #doc.page_path(1, 'page')
                 for i in range(PdfFileReader(open(doc.path(), 'rb')).getNumPages()):
                     todo.append(dict(doc=doc, page=i+1, lang=lang, ocr_resolution=ocr_resolution, psm=psm, x=x, y=y, W=W, H=H, pdf_to_ppm=pdf_to_ppm, user_code=user_code, user=user, pdf=pdf, preserve_color=preserve_color))
+            elif doc.extension in ("docx", "doc", "odt", "rtf"):
+                import docassemble.base.util
+                doc_conv = docassemble.base.util.pdf_concatenate(doc)
+                for i in range(PdfFileReader(open(doc_conv.path(), 'rb')).getNumPages()):
+                    todo.append(dict(doc=doc_conv, page=i+1, lang=lang, ocr_resolution=ocr_resolution, psm=psm, x=x, y=y, W=W, H=H, pdf_to_ppm=pdf_to_ppm, user_code=user_code, user=user, pdf=pdf, preserve_color=preserve_color))
             else:
                 todo.append(dict(doc=doc, page=None, lang=lang, ocr_resolution=ocr_resolution, psm=psm, x=x, y=y, W=W, H=H, pdf_to_ppm=pdf_to_ppm, user_code=user_code, user=user, pdf=pdf, preserve_color=preserve_color))
     #sys.stderr.write("ocr_page_tasks finished\n")
@@ -184,82 +198,96 @@ def ocr_pdf(*pargs, target=None, filename=None, lang=None, psm=6, dafilelist=Non
     else:
         device = 'tiffgray'
     docs = []
-    all_pdf = True
     if not isinstance(target, DAFile):
         raise DAError("ocr_pdf: target must be a DAFile")
     for other_file in pargs:
         if isinstance(other_file, DAFileList):
             for other_file_sub in other_file.elements:
-                if not other_file._is_pdf():
-                    all_pdf = False
                 docs.append(other_file_sub)
         elif isinstance(other_file, DAFileCollection):
-            if not hasattr(other_file, 'pdf'):
-                raise DAError('ocr_pdf: DAFileCollection object did not have pdf attribute.')
-            docs.append(other_file.pdf)
+            if hasattr(other_file, 'pdf'):
+                docs.append(other_file.pdf)
+            elif hasattr(other_file, 'docx'):
+                docs.append(other_file.docx)
+            else:
+                raise DAError('ocr_pdf: DAFileCollection object did not have pdf or docx attribute.')
         elif isinstance(other_file, DAStaticFile):
-            if not other_file._is_pdf():
-                all_pdf = False
             docs.append(other_file)
         elif isinstance(other_file, (str, DAFile)):
-            all_pdf = False
             docs.append(other_file)
     if len(docs) == 0:
-        if not target._is_pdf():
-            all_pdf = False
         docs.append(target)
-    if len(docs) > 1 or not all_pdf:
-        import docassemble.base.util
-        doc = docassemble.base.util.pdf_concatenate(docs)
-    else:
-        doc = docs[0]
     if psm is None:
         psm = 6
+    output = []
+    for doc in docs:
+        if not hasattr(doc, 'extension'):
+            continue
+        if doc._is_pdf() and hasattr(doc, 'has_ocr') and doc.has_ocr:
+            output.append(doc.path())
+            continue
+        if doc.extension in ['png', 'jpg', 'gif']:
+            import docassemble.base.util
+            doc = docassemble.base.util.pdf_concatenate(doc)
+        elif doc.extension in ['docx', 'doc', 'odt', 'rtf']:
+            import docassemble.base.util
+            output.append(docassemble.base.util.pdf_concatenate(doc).path())
+            continue
+        elif not doc._is_pdf():
+            logmessage("ocr_pdf: not a readable image file")
+            continue
+        path = doc.path()
+        pdf_file = tempfile.NamedTemporaryFile(prefix="datemp", mode="wb", delete=False)
+        pdf_file.close()
+        if doc.extension == 'pdf':
+            tiff_file = tempfile.NamedTemporaryFile(prefix="datemp", mode="wb", suffix=".tiff", delete=False)
+            params = ['gs', '-q', '-dNOPAUSE', '-sDEVICE=' + device, '-r600', '-sOutputFile=' + tiff_file.name, path, '-c', 'quit']
+            try:
+                result = subprocess.run(params, timeout=60*60).returncode
+            except subprocess.TimeoutExpired:
+                result = 1
+                logmessage("ocr_pdf: call to gs took too long")
+            if result != 0:
+                raise Exception("ocr_pdf: failed to run gs with command " + " ".join(params))
+            params = ['tesseract', tiff_file.name, pdf_file.name, '-l', str(lang), '--psm', str(psm), '--dpi', '600', 'pdf']
+            try:
+                result = subprocess.run(params, timeout=60*60).returncode
+            except subprocess.TimeoutExpired:
+                result = 1
+                logmessage("ocr_pdf: call to tesseract took too long")
+            if result != 0:
+                raise Exception("ocr_pdf: failed to run tesseract with command " + " ".join(params))
+        else:
+            params = ['tesseract', path, pdf_file.name, '-l', str(lang), '--psm', str(psm), '--dpi', '300', 'pdf']
+            try:
+                result = subprocess.run(params, timeout=60*60).returncode
+            except subprocess.TimeoutExpired:
+                result = 1
+                logmessage("ocr_pdf: call to tesseract took too long")
+            if result != 0:
+                raise Exception("ocr_pdf: failed to run tesseract with command " + " ".join(params))
+        output.append(pdf_file.name + '.pdf')
+    if len(output) == 0:
+        return None
+    if len(output) == 1:
+        the_file = tempfile.NamedTemporaryFile(prefix="datemp", mode="wb", delete=False)
+        the_file.close()
+        shutil.copyfile(output[0], the_file.name)
+        source_file = the_file.name
+    else:
+        import docassemble.base.pandoc
+        source_file = docassemble.base.pandoc.concatenate_files(output)
     if filename is None:
         filename = 'file.pdf'
-    if not hasattr(doc, 'extension'):
-        return None
-    if doc.extension not in ['pdf', 'png', 'jpg', 'gif']:
-        raise DAError("ocr_pdf: not a readable image file")
-    path = doc.path()
-    pdf_file = tempfile.NamedTemporaryFile(prefix="datemp", mode="wb", delete=False)
-    pdf_file.close()
-    if doc.extension == 'pdf':
-        tiff_file = tempfile.NamedTemporaryFile(prefix="datemp", mode="wb", suffix=".tiff", delete=False)
-        params = ['gs', '-q', '-dNOPAUSE', '-sDEVICE=' + device, '-r600', '-sOutputFile=' + tiff_file.name, path, '-c', 'quit']
-        try:
-            result = subprocess.run(params, timeout=60*60).returncode
-        except subprocess.TimeoutExpired:
-            result = 1
-            logmessage("ocr_pdf: call to gs took too long")
-        if result != 0:
-            raise Exception("ocr_pdf: failed to run gs with command " + " ".join(params))
-        params = ['tesseract', tiff_file.name, pdf_file.name, '-l', str(lang), '--psm', str(psm), 'pdf']
-        try:
-            result = subprocess.run(params, timeout=60*60).returncode
-        except subprocess.TimeoutExpired:
-            result = 1
-            logmessage("ocr_pdf: call to tesseract took too long")
-        if result != 0:
-            raise Exception("ocr_pdf: failed to run tesseract with command " + " ".join(params))
-    else:
-        params = ['tesseract', path, pdf_file.name, '-l', str(lang), '--psm', str(psm), 'pdf']
-        try:
-            result = subprocess.run(params, timeout=60*60).returncode
-        except subprocess.TimeoutExpired:
-            result = 1
-            logmessage("ocr_pdf: call to tesseract took too long")
-        if result != 0:
-            raise Exception("ocr_pdf: failed to run tesseract with command " + " ".join(params))
     target.initialize(filename=filename, extension='pdf', mimetype='application/pdf', reinitialize=True)
-    shutil.copyfile(pdf_file.name + '.pdf', target.file_info['path'])
+    shutil.copyfile(source_file, target.file_info['path'])
     del target.file_info
     target._make_pdf_thumbnail(1, both_formats=True)
     target.commit()
     target.retrieve()
     return target
 
-def ocr_page(doc=None, lang=None, pdf_to_ppm='pdf_to_ppm', ocr_resolution=300, psm=6, page=None, x=None, y=None, W=None, H=None, user_code=None, user=None, pdf=False, preserve_color=False):
+def ocr_page(indexno, doc=None, lang=None, pdf_to_ppm='pdf_to_ppm', ocr_resolution=300, psm=6, page=None, x=None, y=None, W=None, H=None, user_code=None, user=None, pdf=False, preserve_color=False):
     """Runs optical character recognition on an image or a page of a PDF file and returns the recognized text."""
     if page is None:
         page = 1
@@ -311,11 +339,11 @@ def ocr_page(doc=None, lang=None, pdf_to_ppm='pdf_to_ppm', ocr_resolution=300, p
         contrast = ImageEnhance.Contrast(brightened)
         final_image = contrast.enhance(2.0)
         file_to_read = tempfile.TemporaryFile()
-        final_image.save(file_to_read, "PNG")
+        final_image.convert('RGBA').save(file_to_read, "PNG")
     file_to_read.seek(0)
     if pdf:
         outfile = doc._pdf_page_path(page)
-        params = ['tesseract', 'stdin', re.sub(r'\.pdf$', '', outfile), '-l', str(lang), '--psm', str(psm), 'pdf']
+        params = ['tesseract', 'stdin', re.sub(r'\.pdf$', '', outfile), '-l', str(lang), '--psm', str(psm), '--dpi', str(ocr_resolution), 'pdf']
         sys.stderr.write("ocr_page: piping to command " + " ".join(params) + "\n")
         try:
             text = subprocess.check_output(params, stdin=file_to_read).decode()
@@ -323,7 +351,7 @@ def ocr_page(doc=None, lang=None, pdf_to_ppm='pdf_to_ppm', ocr_resolution=300, p
             raise Exception("ocr_page: failed to run tesseract with command " + " ".join(params) + ": " + str(err) + " " + str(err.output.decode()))
         sys.stderr.write("ocr_page finished with pdf page " + str(page) + "\n")
         doc.commit()
-        return dict(page=page, doc=doc)
+        return dict(indexno=indexno, page=page, doc=doc)
     params = ['tesseract', 'stdin', 'stdout', '-l', str(lang), '--psm', str(psm)]
     sys.stderr.write("ocr_page: piping to command " + " ".join(params) + "\n")
     try:
@@ -331,4 +359,4 @@ def ocr_page(doc=None, lang=None, pdf_to_ppm='pdf_to_ppm', ocr_resolution=300, p
     except subprocess.CalledProcessError as err:
         raise Exception("ocr_page: failed to run tesseract with command " + " ".join(params) + ": " + str(err) + " " + str(err.output.decode()))
     sys.stderr.write("ocr_page finished with page " + str(page) + "\n")
-    return dict(page=page, text=text)
+    return dict(indexno=indexno, page=page, text=text)
